<!DOCTYPE html>
<html>
  <title>Colav ApiDoc</title>
  <head>
    <meta http-equiv = "refresh" content = "0; url = {{apidoc_link}}" />
  </head>
  <body>
    <p>Redirecting to <a href={{apidoc_link}}> {{apidoc_link}} </a>.</p>
  </body>
</html>